(* -*- coq-load-path: ("." "../compcert/cfrontend" "../compcert/lib" "../compcert/common" "../compcert/backend" "../compcert/x86") -*- *)


Require Import Utf8 Coqlib.
Require Import Maps.
Require Import Ast.
Require Import Integers.
Require Import Pointers.
Require Import Values.
Require Import Mem.
Require Import Events.
Require Import Globalenvs.
Require Import Op.
Require Import Registers.
Require Import Libtactics.
Require Import INJECT.
Require Import Classical.
Require Import Utils.
Require Import Emem Permissions MemoryMachine TSOMemoryMachine RefinesDefs RefinesUtils.
Require Import RefinesRules AllRefinesRulesAux.

Import Bigstep.

Section S.

  Set Implicit Arguments.

Definition frozen t (d: perm_map) p : Prop :=
  match d p with
    | HiddenBy q => q = t
    | FrozenBy s => Rset.mem t s
  end.

Definition frozen_range t d r : Prop :=
  ∃ p, range_inside (p, Int.repr 4) r ∧ frozen t d p.

Inductive le_frozen t : relation permission :=
| KFH : le_frozen t (HiddenBy t) (HiddenBy t)
| KFO t' q : t' ≠ t → le_frozen t (HiddenBy t') q
| KFF s s' (LEF: Rset.mem t s = Rset.mem t s')
  : le_frozen t (FrozenBy s) (FrozenBy s')
.

Remark le_frozen_refl t : ∀ x, le_frozen t x x.
Proof. intros [t'|s]; vauto.
      case (peq t' t). intros ->. vauto. vauto.
Qed.

Remark le_frozen_frozen t d d' p :
  le_frozen t (d p) (d' p) →
  frozen t d p → frozen t d' p.
Proof. unfold frozen. inversion 1; auto; congruence. Qed.

Remark readable_frozen t d p : readable_perm d p t → frozen t d p.
Proof. unfold frozen. inversion 1. autorw'. Qed.

Remark le_frozen_readable t d d' p :
  le_frozen t (d p) (d' p) →
  (readable_perm d p t -> readable_perm d' p t).
Proof.
  inversion 1; inversion 1; autorw'.
  econstructor; eauto.
Qed.

Unset Elimination Schemes.
Record le t m1 m2 : Prop := LE
{ le_perm : ∀ a, wf_ptr a → frozen t m1.(perms) a →
                 le_frozen t (m1.(perms) a) (m2.(perms) a)
; le_load : ∀ a v, wf_ptr a → frozen t m1.(perms) a →
                   load_ptr Mint32 m1 a = Some v → load_ptr Mint32 m2 a = Some v
; le_alloc: ∀ r k, frozen_range t m1.(perms) r →
                   range_allocated r k m1 → range_allocated r k m2
}.

Lemma lepz (i j : int) : Int.unsigned i + Int.unsigned j <= Int.unsigned i →
                         j = Int.zero.
Proof.
  intros H.
  apply Int.unsigned_eq. simpl. rewrite Zmod_small.
  pose proof (Int.unsigned_range j). intuition.
  generalize Int.modulus_pos; intuition.
Qed.

Lemma le_buf t m1 m2 m1' m2' bi e1 e2 :
  le t m1 m2 →
  buf_step t m1 bi e1 m1' →
  buf_step t m2 bi e2 m2' →
  le t m1' m2'.
Proof.
Hint Resolve wf_ptr_aligned ranges_disjoint_comm lepz.
  intros [P L M].
  intros H; inv H; (intros H; inv H); split; simpl; auto.
  > intros a v' A F H.
    case (eq_dec p a). intros ->.
      pose proof (load_store_new _ _ _ _ ESWF2 ESWF1 ESS). autorw'.
      eapply load_store_new; eauto.
    intros K.
    rewrite (load_store_old _ _ _ _ _ ESWF2 ESWF1 A ESS K) in H.
    pose proof (load_store_old _ _ _ _ _ ESWF2 ESWF1 A ESS0 K). autorw'. eauto.
  > simpl in *. intros r k R F.
    pose proof (store_preserves_allocated_ranges ESS r k).
    pose proof (store_preserves_allocated_ranges ESS0 r k).
    intuition.
  > destruct ESF as (wfr & ESF).
    intros a v A F H.
      case (wf_range_ptr_inside wfr A); intros R.
        rewrite (load_alloc_inside ESL) in H; auto. clarify.
        apply (load_alloc_inside ESL0); auto.
      rewrite (load_alloc_other ESL) in H; auto.
      rewrite (load_alloc_other ESL0); eauto.
  > simpl in *. intros r q F Hr.
    destruct (alloc_preserves_allocated_back ESL r q Hr) as [(-> & ->)|K].
      eapply (alloc_someD ESL0).
    apply (alloc_preserves_allocated ESL0). intuition.
  > intros a v A F H.
    assert (ranges_disjoint (range_of_chunk a Mint32) (p, n)) as RD.
     case (wf_range_ptr_inside (proj1 ESWF) A); auto.
     intros K.
     assert (load_ptr Mint32 m' a = None).
     apply (load_free_overlap ESL ESA).
     apply ranges_overlap_comm.
     eapply overlap_inside_overlap. eapply ranges_overlap_refl. vauto. easy.
     now autorw'.

   assert (load_ptr Mint32 m a = Some v) as Q.
     rewrite <- H.
     symmetry; eapply load_free_other; eauto.
   destruct (load_chunk_allocated_someD Q) as ((r & mk & HR & HR') & AL).
   assert (frozen_range t d r) as FP.
     exists a. intuition.
   pose proof (M _ _ FP HR') as Q'. simpl in Q'.
   destruct (ranges_are_disjoint Q' ESA0) as [(? & ?)|K']; clarify.
   destruct (ranges_are_disjoint HR' ESA) as [(? & ?)|K']; clarify.
     pose proof (disjoint_inside_disjoint_r _ _ _ RD HR) as X.
       destruct a. simpl in X. unfold range_of_chunk in X. simpl in X. inv X; clarify.
       assert (Int.repr 4 = Int.zero) by intuition eauto. clarify.
     destruct p. destruct K' as [K'|[K'|K']]. congruence.
     assert (n0 = Int.zero) by eauto. subst.
     pose proof (valid_alloc_range_non_empty (allocated_range_valid _ _ _ ESA0)) as X; inv X.
     assert (n = Int.zero) by eauto. subst.
     pose proof (valid_alloc_range_non_empty (allocated_range_valid _ _ _ ESA)) as X; inv X.
   erewrite (load_free_other ESL0); eauto.
   eapply disjoint_inside_disjoint_l; eauto.

   intros r ok FR RA'.
   pose proof (free_preserves_allocated_back ESL _ _ RA') as RA.
   pose proof (M _ _ FR RA) as RA0.
   case (free_preserves_allocated ESL0 _ _ RA0); trivial.
   destruct r as (r & sz). simpl.
   intros (<- & ->). apply False_ind.
   case (ranges_are_disjoint ESA RA).
   intros (A & B); clarify.
   exact (free_not_allocated ESL _ _ RA').
   destruct r as (r & ri). intros [K|[K|K]]; intuition.
   assert (n = Int.zero) by eauto. subst.
   pose proof (valid_alloc_range_non_empty (allocated_range_valid _ _ _ ESA)) as X; inv X.
   assert (sz = Int.zero) by eauto. subst.
   pose proof (valid_alloc_range_non_empty (allocated_range_valid _ _ _ RA)) as X; inv X.
Qed.

Lemma le_mem_step t t' (NE: t ≠ t') m1 e m2 :
  mem_step t' m1 e m2 →
  le t m1 m2.
Proof.
  Hint Resolve le_frozen_refl.
  intros H; inv H. inv ESD.
  > vauto.
  > vauto.
  > econstructor; eauto.
    unfold has_store_perm, frozen in *. simpl.
    intros a WA FA.
    case (eq_dec a p). intros ->. destruct (d p). autorw'. apply False_ind.
      destruct (Rset.subset_singleton _ _ ESD0); subst.
      eelim (Rset.empty_not_mem); eassumption.
      elim NE. eapply Rset.mem_singleton. eassumption.
      intros. rewrite perm_gso; auto.
  > econstructor; eauto. simpl.
    intros a WA FA. inv ESD0.
    case (eq_dec a p). intros ->. destruct (d p); autorw'. rewrite perm_gss. econstructor.
    symmetry. now eapply Rset.add_mem_o.
    intros n. rewrite perm_gso; eauto.
  > constructor; auto.
    apply release_func in ESP. unfold frozen. simpl.
    intros a WA FA. subst.
    destruct (d a); autorw'. repeat rewrite peq_false; auto.
    econstructor. symmetry. now eapply Rset.mem_rmo.
  > econstructor; eauto.
  > econstructor; unfold frozen; simpl.
    destruct released.
    apply release_func in ESP. subst.
    intros a WA FA.
    case (eq_dec a p). intros ->. autorw'.
    intros n. destruct (d a). subst. repeat rewrite peq_false; auto.
    econstructor. symmetry. now eapply Rset.mem_rmo.
    subst. auto.
    intros a v' WA FA old.
    rewrite (load_store_old _ _ _ _ _ ESWF2 ESWF1 WA ESS). easy.
    intros ->. autorw'.
    intros r k H0 H1. now apply (store_preserves_allocated_ranges ESS).
  > econstructor; eauto.
  > econstructor; eauto; unfold frozen; simpl. inv ESD.
    destruct released.
    apply release_func in ESP. subst.
    intros a WA FA.
    case (eq_dec a p). intros ->. autorw'. destruct (d p); autorw'.
    rewrite peq_false in ESD0; auto.
    clarify.
    inv ESD0.
    econstructor. symmetry. now eapply Rset.mem_rmo.
    intros n. destruct (d a). subst. repeat rewrite peq_false; auto.
    econstructor. symmetry. now eapply Rset.mem_rmo.
    subst. auto.
    intros a v' WA FA old.
    rewrite (load_store_old _ _ _ _ _ ESWF2 ESWF1 WA ESS). easy.
    intros ->. destruct released.
    apply release_func in ESP. subst. inv ESD. destruct (d a). subst.
    rewrite peq_false in ESD0; auto.
    clarify.
    inv ESD0.
    rewrite <- @Rset.mem_rmo with (v':=t') in FA; auto.
    destruct (Rset.subset_singleton _ _ ALONE) as [H|H];
    rewrite H in FA.
    eapply Rset.empty_not_mem; eassumption.
    eapply Rset.empty_not_mem.
    unfold Rset.singleton in FA. rewrite Rset.add_mem_o in FA; eauto.
    subst. inv ESD. autorw'.
    destruct (Rset.subset_singleton _ _ ALONE) as [-> | ->].
    eapply Rset.empty_not_mem; eassumption.
    unfold Rset.singleton in FA. rewrite Rset.add_mem_o in FA; eauto.
    eapply Rset.empty_not_mem; eassumption.
    intros r k H0 H1. now apply (store_preserves_allocated_ranges ESS).
  > econstructor; eauto.
  > assert (∀ a, frozen t d a → p ≠ a) as PA.
      unfold frozen; simpl.
      intros a FA ->.
      destruct r.
        apply release_func in ESP. subst.
        destruct (d a). subst.
        rewrite peq_false in ESD; auto.
        clarify.
        elim (@Rset.empty_not_mem t). replace Rset.empty with (Rset.remove t' s).
        rewrite Rset.mem_rmo; auto. unfold empty_perm in *. congruence.
      subst. destruct (d' a); clarify.
      unfold empty_perm in ESD. replace s with Rset.empty in FA by congruence.
      elim (Rset.empty_not_mem FA).
    econstructor; eauto; unfold frozen; simpl.
    destruct r. 2: subst; auto.
    apply release_func in ESP. subst.
    intros a WA FA.
    case (eq_dec a p). intros ->. eelim PA; eauto.
    intros n. destruct (d a). subst. repeat rewrite peq_false; auto.
    econstructor. symmetry. now eapply Rset.mem_rmo.
    intros a v' WA FA old.
    destruct instr; clarify. simpl in STO.
    destruct ESWF3 as [Hold Hnew].
    destruct (Val.eq_dec). clarify.
    destruct (Val.eq_dec). clarify.
    simpl in STO.
    pose proof (PA _ FA).
    destruct (Val.eq_dec). clarify.
    now rewrite (load_store_old _ _ _ _ _ ESWF2 Hnew WA STO).
    now rewrite (load_store_old _ _ _ _ _ ESWF2 ESWF1 WA STO).
    intros r' k H0 H1. now apply (store_preserves_allocated_ranges STO).
  > econstructor; eauto.
    intros a v WA FA old.
    destruct ESF as [wfr ESF].
    case (wf_range_ptr_inside wfr WA).
    intros. apply False_ind.
    destruct (load_chunk_allocated_someD old) as [(r' & ok & Ri & Ra) H].
    eapply (alloc_someD ESL). 2: eapply Ra.
    eapply overlap_inside_overlap. 2: eassumption.
    apply ranges_overlap_comm.
    eapply overlap_inside_overlap. 2: eassumption.
    apply ranges_overlap_refl. vauto.
    intros; rewrite (load_alloc_other ESL); auto.
    now intros; apply (alloc_preserves_allocated ESL).
  > econstructor; eauto.
    destruct ESWF as (wfr & K).
    simpl. intros a v WA FA old.
    case (wf_range_ptr_inside wfr WA).
    intros Q. apply False_ind.
    generalize (K _ Q). unfold has_store_perm. unfold frozen in FA.
    destruct (d a). intros ->. clarify.
    intros ALONE.
    destruct (Rset.subset_singleton _ _ ALONE) as [H|H];
    rewrite H in FA.
    eapply Rset.empty_not_mem; eassumption.
    eapply Rset.empty_not_mem.
    unfold Rset.singleton in FA. rewrite Rset.add_mem_o in FA; eauto.
    intros Q. rewrite <- old. eapply (load_free_other ESL). eassumption. auto.
    simpl. intros r ok (q & Q & FQ) old.
    destruct (free_preserves_allocated ESL _ _ old) as [H|[X Y]]. easy.
    apply False_ind.
    destruct r as (x & sz). subst.
    pose proof (alloc_ranges_same old ESA) as [? ?]. subst.
    destruct ESWF as (WX & K).
    generalize (K _ Q). unfold has_store_perm. unfold frozen in FQ.
    destruct (d q). intros ->. clarify.
    intros ALONE.
    destruct (Rset.subset_singleton _ _ ALONE) as [H|H];
    rewrite H in FQ.
    eapply Rset.empty_not_mem; eassumption.
    eapply Rset.empty_not_mem.
    unfold Rset.singleton in FQ. rewrite Rset.add_mem_o in FQ; eauto.
Qed.

End S.

